import click
from .cli import add, config, _list, edit, balance #, delete
from .utils.data_utils import ensure_data_files
import art


# The Base of CLI Application
# app() groups in itself many subcommands 
# Making the cli structure look like:
# App 
#   |_ balance
#   |_ add
#   |_ edit
#   |_ config
#   |_ search
##   |_delete 
# 
# Each command located in a separate {command}.py file 


@click.group(invoke_without_command=True) # Initing Click Group with the ability of calling the app without any subcommand
@click.pass_context # Getting access to applications context
def app(ctx):
    
    # This command checks database file existance on each command call
    # in case the files do not exist, it will create them
    ensure_data_files() 
    # If application called without invoking subcommands
    if ctx.invoked_subcommand is None:
        # It'll print Application name as green ASCII Art  
        click.secho(art.text2art("FINMAN", space=1), fg='green')
#         _____   ___   _   _   __  __      _      _   _ 
#        |  ___| |_ _| | \ | | |  \/  |    / \    | \ | |
#        | |_     | |  |  \| | | |\/| |   / _ \   |  \| |
#        |  _|    | |  | |\  | | |  | |  / ___ \  | |\  |
#        |_|     |___| |_| \_| |_|  |_| /_/   \_\ |_| \_|
#                                                               
    # Otherwise, it'll pass the execution to the proper subcommand 
    else:
        pass
    
# Adding subcommands to application group
app.add_command(add)
app.add_command(_list)
app.add_command(config)
app.add_command(edit)
app.add_command(balance)
# app.add_command(delete) # This command will be added in future


# P.S. About the ASCII Art: Inspired by Kali Linux Terminal Applications


import click
from rich.prompt import Prompt
from ..utils.model_utils import DATE_NOW
from ..utils.formatter import table_view
from ..utils.data_utils import save_records, get_wallet_by_name, update_wallet
from ..utils.config_utils import get_config_value
from ..models import Record, Category, Wallet

# Add Command 
# used for adding a Record to JSON DataBase and updating Wallet's data accordingly
# U can add a record by specifing at least two things which are Category and Amount of Money

# Category does have a default value, it is an Expense Category [-]

# While amount doesn't have a directly default value, but u still can add a Record with 0.0 amount of money

# The discription will be prompted too, but the default will be ''
# so by pressing enter u can add an empty description message

# Default date value is datetime.now() 'Todays Date'


@click.command("add",
               help="Add a new record")
@click.option("--date", "-d",
              help="Record date YYYY-MM-DD",
              required=False,
              type=str,
              default=DATE_NOW) # Todays date
@click.option("--category", "-c", 
              help="Record category", 
              required=False, 
              default="-",                  # Default value of category is Expense
              show_default=True,
              type=click.Choice([category.value for category in Category]), # choose from Category(Enum) object values
              prompt="Record Category")
@click.option("--amount", "-a",
              help="Amount of money",
              required=False,
              type=float,
              default=0.0)                  # Amount's default is 0.0
@click.option("--message", "-m",
              help="Description message of a record",
              required=False,
              type=str,
              default='',
              prompt="Add record description")
# This option will be added in the future for managing many wallet's records
# @click.option("--wallet", "-w",
#               help="Defines wallet (to be added...)",
#               required=False,
#               type=str,
#               default=get_config_value('default_wallet'))
def add(date, category, amount, message):
    """Adds a Record, stores it in JSon DataBase and updates Wallet's data

    Args:
        date (str: YYYY-MM-DD): Records date, default to Today's Date
        category (str: [+ | -]): Income | Expense:Default 
        amount (float): Amount of Money
        message (str): A short description message about the Record
    """
    # Starting an inifint loop to gather all info about the record
    # in case if some options not passed directly with the command
    while True:  
        # Main one is : Amount of money, to avoid 0.0$ Record
        if not amount:
            amount = Prompt.ask("Amount of money :dollar:")
        
        # Initializing a Record Object with passed data    
        record = Record(date, Category(category).en_text, amount, message)
        # Printing the record in table view to show the data clearly [example below]
        click.echo(f"{table_view(record)}\n")
        # Styling a confirmation prompt
        confirmation_text = click.style("Save this record?", fg='yellow', bold=True)
        # Prompting confirmation for saving Records Data
        
        # Default = [YES]
        # press Enter to go 
        # or type [no] to aboard
        confirmation = click.confirm(confirmation_text, default=True)
        # If Confirmed
        if confirmation:
            # Saving record to JSON Database
            save_records(record)
            # Updating Wallet's data based on saved record
            update_wallet(get_config_value("default_wallet"), record)
            # Styling a Record ID to display it  
            record_id = click.style(f'{record._id}', bold=True, fg="cyan")
            saved_message_text = click.style('Saved Successfully!', fg='green')
            # Displaying successful record saving message
            click.secho(f"Record {record_id} {saved_message_text}", fg='green')
            
            #Breaking loop
            break
# End of command execution
        # In case the confiramtion declined
        else:
            # Prompt the user asking him if he want to add Record again
            try_again_text = click.style("Do you want to enter the record again?", fg='yellow')
            try_again = click.confirm(try_again_text, default=False)
            
            # Default = NO
            # press Enter to aboard and end command execution
            # or type [yes] to procced with adding record again
            
            # if Declined [pressed enter | typed no]
            if not try_again:
                # Displaying message about cancelation
                click.secho("Record entry has been cancelled.", fg='red')
                break
            #End of command execution
            
            # If agreed [typed yes]
            else:
                # The application will prompt the user with questions to gather information about Record 
                date = click.prompt("Enter records date", default=DATE_NOW)
                category = click.prompt("You Got or Paid",
                                        type=click.Choice([category.value for category in Category]),
                                        default="-",
                                        show_default=False)
                amount = Prompt.ask("How much money :dollar:")
                message = click.prompt("Record Description", default='')
            # In this case you can exit either by pressing [Ctrl+C] or by answering the next confirmation question with [NO]
   
   # Example of Record's table view             
# ╭────────────┬────────────┬──────────┬───────────────╮
# │ Date       │ Category   │   Amount │ Description   │
# ├────────────┼────────────┼──────────┼───────────────┤
# │ 2024-05-09 │ Income     │     5000 │ Pop-corn      │
# ╰────────────┴────────────┴──────────┴───────────────╯

# Save this record? [Y/n]: 
# Record 663cda40 Saved Successfully!

# P.S.: Yes, this is the minimal price of pop-corn in Syria, tbh, it's only ~0.3 $ LOL


# Main data manipulation utilities and functions 
from typing import Dict, List
from pathlib import Path
import json
import os
from finman.config.config import DATA_DIR, RECORDS_DATA_FILE, WALLETS_DATA_FILE
from .config_utils import update_config
from .formatter import format_str
from finman.models import Record, Wallet
import click
import art
from time import sleep


# This function runs only if the wallets.json not exist
def first_run() -> Wallet:
    """Prompting wallet name from user on first run or when wallet.json file deleted

    Returns:
        Wallet: Wallet object representing the wallet for further record adding
    """
    # Priniting ASCII art app name 
    # Prompting a name for the wallet 
    # Initializing wallet object with given name
    click.secho("\nWelcome to ", fg="green")
    click.secho(art.text2art("FINMAN", space=1), fg="red")
    click.secho("Please enter a name for your wallet\n", fg="green")
    name = click.prompt(format_str("You can change it later", "green"), type=str, default="My_wallet")
    return Wallet(name=name)


def ensure_data_files() -> None:
    """Checking json databses within databse directorty, creates them if not exist
    """
    # Ensure the data directory exists
    os.makedirs(DATA_DIR, exist_ok=True)  
    if not os.path.exists(RECORDS_DATA_FILE):
        with open(RECORDS_DATA_FILE, 'w') as file:
            # Create an empty dict for records
            json.dump({}, file)  
    if not os.path.exists(WALLETS_DATA_FILE):
        with open(WALLETS_DATA_FILE, 'w') as file:
            data = {}
            wallet = first_run()
            data[wallet.name] = wallet._jsonfy()
            sleep(0.5)
            update_config('default_wallet', wallet.name)
            json.dump(data, file, indent=4)


def load_wallets(filename: str) -> List:
    ensure_data_files()
    wallets = []
    with open(filename, 'r') as f:
        loaded_wallets = json.load(f)
        for name, wallet in loaded_wallets.items():
            wallets.append(wallet)
    return wallets


def get_wallet_by_name(wallet_name: str) -> Wallet:
    wallets = load_wallets(WALLETS_DATA_FILE)
    for _ in wallets:
        if _["name"] == wallet_name:
            wallet_data = Wallet._unjsonfy(_)
            return Wallet(wallet_data[0], wallet_data[1], wallet_data[2], wallet_data[3])
    return None

def update_wallet(wallet_name: str, record: Record) -> None:
    data = {}
    
    with open(WALLETS_DATA_FILE, 'r') as file:
        data = json.load(file)
    
    wallet = get_wallet_by_name(wallet_name)
    wallet.update_data(record)
    
    data[wallet.name] = wallet._jsonfy()
    
    with open(WALLETS_DATA_FILE, 'w') as file:
        json.dump(data, file, indent=4)
        

def load_records(filename: str) -> List:
    ensure_data_files()
    records = []
    with open(filename, 'r') as f:
        loaded_data = json.load(f)
        for id, record in loaded_data.items():
            records.append(record)
    return records


def load_records_extended(filename: str) -> List:
    ensure_data_files()
    records = []
    with open(filename, 'r') as f:
        loaded_data = json.load(f)
        for id, record in loaded_data.items():
            record["id"] = id
            records.append(record)
    return records


def save_records(records):
    data = {}
    with open(RECORDS_DATA_FILE, 'r') as file:
            data = json.load(file)
    # Update the data with new records
    if isinstance(records, list):
        for record in records:
            data[record._id] = record._jsonify()

    if isinstance(records, Record):
        data[records._id] = records._jsonfy()
    
    # Write the updated data back to the file
    with open(RECORDS_DATA_FILE, 'w') as file:
        json.dump(data, file, indent=4)



def search_records(query):
    all_records = load_records_extended(RECORDS_DATA_FILE)
    filtered_records = []
    for record in all_records:
        match = True
        if isinstance(record, dict):
            record_dict = record
        else:
            record_dict = record.__dict__
        for key, value in query.items():
            if key not in record_dict or str(record_dict.get(key)) != str(value):
                match = False
                break
        if match:
            filtered_records.append(record)
    return filtered_records


def get_record_by_id(_id: str) -> Dict:
    loaded_data = load_records_extended(RECORDS_DATA_FILE)
    for record in loaded_data:
        if record["id"] == _id:
            return record
    return None

def update_record(_id: str, updated_record: Record):
    data = {}
    with open(RECORDS_DATA_FILE, 'r') as file:
            data = json.load(file)
    
    data[_id] = updated_record._jsonfy()
    
    with open(RECORDS_DATA_FILE, 'w') as file:
        json.dump(data, file, indent=4)
    


from pathlib import Path
# File with constants e.g. pathes to JSON database and application folders

MAIN_DIR = Path(__file__).resolve().parent.parent
DATA_DIR = MAIN_DIR.joinpath("database")

CONFIG_FILE=MAIN_DIR.joinpath("config", "config.json")
RECORDS_DATA_FILE=DATA_DIR.joinpath("records.json")
WALLETS_DATA_FILE=DATA_DIR.joinpath("wallets.json")



from enum import Enum

# Category class inherited from Enum calss 
# To make it easier managing the category type of a record 
# making each category accessible with many formats 

class Category(Enum):
    def __new__(cls, symbol, code, ru_text, en_text):
        obj = object.__new__(cls)
        obj._value_ = symbol 
        obj.code = code
        obj.ru_text = ru_text
        obj.en_text = en_text
        return obj

    INCOME = ("+", "1", "Доход", "Income")
    EXPENSE = ("-", "0", "Расход", "Expense")





from typing import List, Dict
from .record import Record
from .enums import Category

# Wallet class representing the base bank 
# Further versions will have ability of dealing with many wallets 
class Wallet:
    """
    Wallet class representing the base bank
    """ 
       
    def __init__(self, name: str, balance: float = 0.0, income: float = 0.0, expense: float = 0.0):
        """Intializes Wallet object by passing name [required]

        Args:
            name (str): Wallet Name 
            balance (float, optional): Base Balance  Defaults to 0.0.
            income (float, optional): Total Incomes Defaults to 0.0.
            expense (float, optional): Total Expenses Defaults to 0.0.
        """
        self.name = name 
        self.balance = balance
        self.income = income
        self.expense = expense
        
    
    # A function for getting wallets data in list format
    # Similar to Record._data()    
    def _data(self) -> List:
        """
        Returns:
            List: with wallets main data 
        """
        return [self.name, self.balance, self.income, self.expense]
    
    # Similar to Record._unjsonfy() method
    @staticmethod    
    def _unjsonfy(data: Dict) -> List:
        """Getting Dict of Wallet's data returns it in list format

        Args:
            data (Dict): Wallet's data 

        Raises:
            ValueError: If passed Dict misses some main Wallet data, it'll raise a ValueError

        Returns:
            List: Wallet's data in list format [name, balance, income, expense]
        """
        if isinstance(data, dict) and all(key in data for key in ['name', 'balance', 'income', 'expense']):
            return [data['name'], data['balance'], data['income'], data['expense']]
        else:
            raise ValueError("Invalid data format for unjsonfy method")
    
    
    # Similar to Record._jsonfy() method   
    def _jsonfy(self) -> Dict:
        """
        Returns:
            Dict: Returns Wallet's data in dictionary 
        """
        return {
            "name": self.name,
            "balance": self.balance,
            "income": self.income,
            "expense": self.expense
        }
        
    def update_data(self, record: Record) -> None:
        """Updating wallet's balance, income and expense based on given record

        Args:
            record (Record): Record to be added to the wallet
        """
        # Checking the category type for adding 
        # Income
        if record.category is Category.INCOME.en_text:
            self.balance += record.amount
            self.income += record.amount
        # Expense
        if record.category is Category.EXPENSE.en_text:
            self.balance -= record.amount
            self.expense -= record.amount
            
        
            



# Importing file paths
from ..config.config import * 
# Importing for combining utils
from .model_utils import *
from .formatter import *
from .cli_utils import *
from .data_utils import *
from .config_utils import *





from datetime import datetime

DATE_NOW = datetime.now().strftime("%Y-%m-%d")

def generate_hex_id() -> str:
    # Get the current date and time
    now = datetime.now()
    # Convert datetime to a timestamp (number of seconds since epoch)
    timestamp = int(now.timestamp())
    # Convert the timestamp to hexadecimal
    hex_id = hex(timestamp)[2:]  # [2:] to strip the '0x' prefix
    return hex_id


     




from typing import List, Any, Dict
from ..models import Record, Wallet
from tabulate import tabulate
import click

# Function for creating table view of a record an printing it
# Depends on Tabulate library
def table_view(record: Any) -> str:
    """Creating a Table view of a record

    Args:
        record (Record | List[Dict | Records | List] | Dict): Record object refers to a invoice record

    Raises:
        ValueError: The function do not accept objects other than Record

    Returns:
        str: String formatted table to be used in terminal
    """
    # Defining Table header
    header = ["Date", "Category", "Amount", "Description"]
    # Checking the inputed type
    if isinstance(record, Record):
        # Getting Record Data
        data = [record._data()]
        
        # Return it as a table in string format
        return tabulate(data, headers=header, tablefmt='rounded_grid')
    # If the input was a list
    if isinstance(record, List):
        # Based of list's data types
        # If list of dictionaries
        if isinstance(record[0], Dict):
            records = []
            for _ in record:
                # Convert the Dict to list of data
                records.append(Record._unjsonfy(_))
            
            return tabulate(records, headers=header, tablefmt='rounded_grid')
        # If the list of Records
        elif isinstance(record[0], Record):
            # Get records data
            data = []
            for _ in record:
                data.append(record._data())
                
            return tabulate(data, headers=header, tablefmt='rounded_grid')
        else:
            # Print error message if list not contains values listed above
            raise ValueError(f"table_view function does not accept list of {type(record)} as input value.")
    # if the inputed type was Dictionary
    elif isinstance(record, Dict):
        # Convert it to list of record's data and return it
        return tabulate([Record._unjsonfy(record)], header, tablefmt='rounded_grid')
    else:
        # Otherwise, throw an error
        raise ValueError(f"table_view function does not accept input value of {type(record)}.")


# Same as table view but with additional ID column
def extended_table_view(record: Any = None) -> str:
    """Creating a Table view of a record

    Args:
        record (Record | List[Dict | Records | List] | Dict): Record object refers to a invoice record

    Raises:
        ValueError: The function do not accept objects other than Record

    Returns:
        str: String formatted table to be used in terminal
    """
    header = ["ID", "Date", "Category", "Amount", "Description"]
    if isinstance(record, Record):
        data = [record._data()]
        return tabulate(data, headers=header, tablefmt='rounded_grid')
    elif isinstance(record, List):
        if isinstance(record[0], Dict):
            records = []
            for _ in record:
                records.append(Record._unjsonfy_extended(_))

            return tabulate(records, headers=header, tablefmt='rounded_grid')
        else:
            return tabulate(record, headers=header, tablefmt='rounded_grid')
    elif isinstance(record, Dict):
        return tabulate([Record._unjsonfy_extended(record)], header, tablefmt='rounded_grid')
    else:
        raise ValueError(f"table_view function does not accept input value of {type(record)}")
    
# Table that shows up when editing a record compairing the original and edited tables    
def updater_table(original_record: Record, updated_record: Record) -> str:
    """A Table that compares an original record with edited (new values) record

    Args:
        original_record (Record): Original record to be edited
        updated_record (Record): The new record with new values to be comaperd

    Returns:
        str: String format comparsion table 
    """
    header = ["ID", "Date", "Category", "Amount", "Description"]
    arrow = ["\U0001F53B" for _ in range(4)]
    arrow.insert(0, "\033[92m=\033[0m")
    data = [original_record._data_extended(), arrow, updated_record._data_extended()]
    return tabulate(data, header, tablefmt="rounded_grid")

# A simplify version of click.style() function for text decoration that based on it
# Accepts bold and italic as **kwargs
# For more customization use original click.style()
def format_str(str: str, color: str, **kwargs) -> str:
    """A simplify version of click.style() function for text decoration that based on it

    Args:
        str (str): Input String to be decorated
        color (str): ['red', 'blue', 'yellow', 'green', white', 'cyan', 'magenta]
        and its bright versions as 'color_bright'

    Returns:
        str: Decorated string
    """
    # Managing kwargs if passed
    if kwargs:
        # Bold
        if "bold" in kwargs.keys() and kwargs["bold"]:
            return click.style(str, fg=color, bold=True)
        # Italic
        if "italic" in kwargs.keys() and kwargs["italic"]:
            return click.style(str, fg=color, italic=True)

    return click.style(str, fg=color)


# Balance table string
def balance_table(wallet: Wallet | List) -> str:
    """Function for creating table view of a Wllet data (will be updated with more functionality)

    Args:
        wallet (Wallet | List): A wallet data in varius typs

    Returns:
        str: Table as a string with needed values
    """
    header = [format_str("Balance", "yellow", bold=True),
              format_str("Income(+)", color="green"),
              format_str("Expense(-)", color="red")]
    if isinstance(wallet, Wallet):
        wallet_name = wallet._data()[0]
        data = [wallet._data()[1:]]
        
        return tabulate(data, header, tablefmt="heavy_grid")
    
    elif isinstance(wallet, List):
        if isinstance(wallet[0], Wallet):
            for _ in (wallet):
                data = [_._data()[1:]]
                
                return tabulate(data, header, tablefmt="heavy_grid")
    else:
        data = wallet[1:]
        
        return tabulate(data, header, tablefmt="heavy_grid")
    



import click
from ..utils.data_utils import get_record_by_id, update_record
from ..utils.formatter import extended_table_view, updater_table
from ..models import Category, Record

# Edit command 
# Gives the ability to edit a Record stored in JSON Databse file 
# by passing it's Unique ID

@click.command("edit",
               help="Edit a Record fetched by it's ID.")
@click.option("--id", "_id",
              help="ID of Record to be edited.",
              type=str,
              required=True)
def edit(_id):
    """Edits a Record saved in JSON Databse file
    takes only one option [ID]

    Args:
        _id (str): Uniqe ID of Record to be edited
    """
    # Prompt the user for an ID if not passed 
    if not _id:
        _id =click.prompt('Enter ID of the Record you want to edit', type=str)
    
    # Fetching Record with given ID 
    record_data = get_record_by_id(_id) # The function will return a Dict
    # Checks if there is a record with given ID
    if not record_data:
        # Prints a log to terminal if Record not found
        click.secho(f"ERROR404: No record found with ID {_id}", fg="red")
    # If there is a Record
    else:
        # Stores it in a Record Object
        original_record = Record(record_data["date"],
                                 record_data["category"],
                                 record_data["amount"],
                                 record_data["message"],
                                 _id) # Passing the given ID to avoid generating new one
        # Prints a table with fetched Record
        click.echo(extended_table_view(record_data))
        # Starts prompting the user to take new data
        while True:
            # Every prompt will ask the required data and show the default (original) one
            date_ = click.prompt("Enter new record date", default=record_data["date"])
            amount_ = click.prompt("Enter new amount", default=record_data["amount"])
            category_ = click.prompt("Enter new Category",
                                        type=click.Choice([category.en_text for category in Category]),
                                        default=record_data['category'])
            message_ = click.prompt("Enter record description message", default=record_data["message"])
            # Creating an updated (edited) version of the Record with same ID
            # Store it in Record Object 
            updated_record = Record(date_, category_, amount_, message_, _id)
            # Printing a comparison table to show the difference between Original and Edited
            click.echo(updater_table(original_record, updated_record))
            # Styles a confirmation prompt
            confirmation_text = click.style("Update record?", fg='yellow', bold=True)
            # Asks for confirmation to edit the Record
            confirmation = click.confirm(confirmation_text, default=True)
            # If confirmation approved
            if confirmation:
                # Updates the Record with new data
                # by replacing the original_record with updated_record
                update_record(_id, updated_record)
                # Prints a succession log
                click.secho("Record updated successfully!", fg='green')
                # Breaks the loop
                break
            # End of command exectution
            
            # If confirmation declined 
            else:
                # Cancle changes and prints a related terminal message
                click.secho("Changes Canceled, Editing Aborted!", fg="red")
                # Breaks the loop
                break
            
# End of Command Execution        


import click
from ..config.config import CONFIG_FILE

@click.command("config",
               help="Edit application configurations")
@click.option("--set", 
              help="Set a configuration option in Key Value format.",
              multiple=True,
              type=(str, str))
@click.option("--get",
              help="Get a configuration option's value") 
def config(set, get):
    config_file = CONFIG_FILE
    click.echo(f"The path of config file is {config_file}")
    


# Configuration file utilities 
from typing import Dict
from ..config.config import CONFIG_FILE
import json
import click



# Fetching configuration file to Dict and returning it
def load_config() -> Dict:
    """Fetching configurations (config.json)

    Returns:
        Dict: congi.json in Dict format
    """
    # Checks configurations file existanse
    # If found
    try:
        # Open in Read mode
        with open(CONFIG_FILE, 'r') as file:
            # Load JSON file format and returns it
            config = json.load(file)
            return config
    # If file not found prints error meassage
    except FileNotFoundError:
        print("Error404: Configuration file 'config.json' not found.")
        

# Updating configuration file with custome configs
def update_config(config_param: str, config_value: str) -> None:
    """Changing Configuration parameter in config.json

    Args:
        config_param (str): The parameter to be changed
        config_value (str): New value
    """
    # Getting the configurations as a Dict
    config = load_config()
    # If the configuration parameter to be edited exists
    if config[config_param]:
        # It'll change it's value to the custom passed value
        config[config_param] = config_value
        # Writes changes to config.JSON
        with open(CONFIG_FILE, 'w') as file:
            json.dump(config, file, indent=4)
    # Otherwise, it will prints that the passed configuration parameter not found
    else:
        click.secho(f"{config_param} not found in configuration file.", fg='red')
        
        # Getting a specific config parameter value       
def get_config_value(config_param: str) -> str:
    """Fetching configuration's parameter value

    Args:
        config_param (str): The configuration parameter 

    Returns:
        str: It's value
    """
    # Loading configurations file 
    config = load_config()
    # If there is a configuration parameter u asking for 
    if config[config_param]:
        # It'll return it's value
        return config[config_param]
    # Otherwise, prints error message
    else:
        click.secho(f"{config_param} not found in configuration file.", fg='red')




import click
from ..utils.formatter import balance_table
from ..utils.data_utils import get_wallet_by_name # ,load_wallets # Will be added in future
from ..utils.config_utils import get_config_value
from ..config.config import WALLETS_DATA_FILE, CONFIG_FILE

# Balance Command 
# no arguments required (yet)
# Shows Wallet data in table view
# Balance, Total Income and Total Expense 



@click.command("balance",
               help="Displays your current Balance, Incomes and Exposes")
def balance():
    """Shows Balance, Total Income and Total Expense of default app wallet
    'Upcoming versions will have some options to work with more than one (default) wallet'
    """
    # Gets the default Wallet name from config.json applications configurations file
    wallet_name = get_config_value("default_wallet")
    # Prints Wallet's data in table view using balance_table() function by passing wallet's name
# ┏━━━━━━━━━━━┳━━━━━━━━━━━━━┳━━━━━━━━━━━━━━┓
# ┃   Balance ┃   Income(+) ┃   Expense(-) ┃
# ┣━━━━━━━━━━━╋━━━━━━━━━━━━━╋━━━━━━━━━━━━━━┫
# ┃         0 ┃           0 ┃            0 ┃
# ┗━━━━━━━━━━━┻━━━━━━━━━━━━━┻━━━━━━━━━━━━━━┛    
    click.echo(balance_table(get_wallet_by_name(wallet_name)))


from typing import List, Dict

# Record class representing the data for each record
class Record:
    """
    Record class representing the data for each record
    """
    def __init__(self, date, category, amount, message, _id = None):
        """
        Initialize a new record

        Args:
            date (str): Records date
            category (str): Records category
            amount (float): Records amount
            message (str): Records description
        """
        # Importin generate_hex_id() to give each record a unique id 
        from ..utils.model_utils import generate_hex_id
        self._id = _id if _id else generate_hex_id()
        self.date = date
        self.category = category
        self.amount = amount
        self.message = message
    
    
    # A static method for converitng a dictionary type data of Record class into a list for further use
    @staticmethod    
    def _unjsonfy(data: Dict) -> List:
        """Converting Dict format Record data into List

        Args:
            data (Dict): Dictionary of Record data (stored in the JSON db as a value of it's _id)
            {
                "date": "date_value",
                "category": "category_value",
                "amount": "amount_value",
                "message": "message_value"}

        Raises:
            ValueError: In case the Dict is incorrect, or misses some main Record Args  

        Returns:
            List: List of data in following format [date, category, amount, message]
        """
        if isinstance(data, dict) and all(key in data for key in ['date', 'category', 'amount', 'message']):
            return [data['date'], data['category'], data['amount'], data['message']]
        else:
            raise ValueError("Invalid data format for unjsonfy method")
    
    
    # Same as previuos method, but extended to have the _id of the Record
    # (Will be depricated in further versions) 
    @staticmethod    
    def _unjsonfy_extended(data: Dict) -> List:
        """Converting Dict format Record data into List

        Args:
            data (Dict): Dictionary of Record data (stored in the JSON db as a value of it's _id)
            {
                "id": "id_value",
                "date": "date_value",
                "category": "category_value",
                "amount": "amount_value",
                "message": "message_value"}

        Raises:
            ValueError: In case the Dict is incorrect, or misses some main Record Args  

        Returns:
            List: List of data in following format [_id, date, category, amount, message]
        """
        if isinstance(data, dict) and all(key in data for key in ['id', 'date', 'category', 'amount', 'message']):
            return [data["id"], data['date'], data['category'], data['amount'], data['message']]
        else:
            raise ValueError("Invalid data format for extended unjsonfy method")
    
    
    # A method for converting a Record to a Dict for storing in JSON DB
    def _jsonfy(self) -> Dict:
        """Converts Record to Dict for easier use with JSON DB file

        Returns:
            Dict: {
                "date": "date_value",
                "category": "category_value",
                "amount": "amount_value",
                "message": "message_value"}
        """
        return {
            "date": self.date,
            "category": self.category,
            "amount": self.amount,
            "message": self.message
        }
    
        
    # A method for manging the data of Record     
    def _data(self) -> List:
        """Returns main Record's data in a List format

        Returns:
            List: [date, category, amount, message]
        """
        return [self.date, self.category, self.amount, self.message]
    
    
    # Same as previuos method, but extended to have the _id of the Record
    # (Will be depricated in further versions) 
    def _data_extended(self) -> List:
        """Returns main Record's data in a List format

        Returns:
            List: [_id, date, category, amount, message]
        """
        return [self._id, self.date, self.category, self.amount, self.message]
    
    


import click
from ..utils.data_utils import search_records
from ..utils.formatter import extended_table_view

# List Command
# Lists Record that meets the given criteria (if passed)

# You can search by any criteria you want, even combine criterias
# Unlike Add command that have an option -c --category that takes category type [Income | Expense]
# Here u can specify the category by passing a flags [--income | --expense] to list one of them

# No defaults (Yet)
# By calling the command with no options it will display all saved records

@click.command("list",
               help="Search for a record")
@click.option("--date", "-d",
              help="Search by Date criteria YYYY-MM-DD",
              required=False,
              type=str)
@click.option("--income", 'category',
              help="Search for Incomes",
              flag_value="Income",
              required=False)
@click.option("--expense", 'category',
              help="Search in Expenses",
              flag_value="Expense",
              required=False)
@click.option("--amount", "-a",
              help="Search by Amount of money",
              required=False,
              type=float)
@click.option("--message", "-m",
              help="Search by description Message",
              required=False,
              type=str)
@click.option("--id",
              help="Search by Record ID criteria.",
              required=False,
              type=str)
def _list(date, category, amount, message, id):
    """Searches for records that meets given criteria(s)

    Args:
        date (str: YYYY-MM-DD): Search by Records Date
        category (flag: [--income | --expense]): Search in Incomes OR Expenses
        amount (float): Search by Amount of Money
        message (str): Search by message description of a record (not part of message)
        id (str): Search by Record's Unique ID (outputs one record if exists) 
    """
    # Creating a query Dict to pack all criterias (if more than one)
    query = {}
    # Declaring a list for displaying criterias with results
    filters = []
    filter_text = ''
    # Checking what criterias were passed with the command 
    if date:
        query['date'] = date
        filters.append(f'Date: {date}')
    if category and isinstance(category, str):
        query['category'] = category
        filters.append(f'Category: {category}')
    if amount:
        query['amount'] = amount
        filters.append(f'Amount: {amount}')
    if message:
        query['message'] = message
        filters.append(f'Discription: {message}')
    if id:
        query['id'] = id
        filters.append(f'Record ID: {id}')

    # Writing filters in one line
    for _ in filters:
        filter_text += f"{_}, "
    # Getting search results of query     
    results = search_records(query)
    # If there is data fetched 
    if results:
        # Print it in a table view
        result_table = extended_table_view(results)
        search_result_text = click.style("Search Results: ", fg="green")
        # Printing filters (criterias) above the table
        click.echo(f"{search_result_text}\nFilters: {filter_text[:-2]}")
        click.echo(result_table)
    # If no data fetched    
    else:
        # display a mesage that no matching records found 
        click.secho("ERROR404:No records found matching the criteria.", fg="red")


